"""Filesystem helpers shared across dataset modules."""
from __future__ import annotations

import tarfile
import time
import urllib.request
from pathlib import Path
from typing import Optional

import jax
import jax.numpy as jnp
import numpy as np


def download_with_progress(url: str, path: Path) -> Path:
    """Download ``url`` to ``path`` while printing percent and window speed."""
    path.parent.mkdir(parents=True, exist_ok=True)
    last_percent = [-1]
    last_bytes = [None]
    last_time = [None]

    def _reporthook(block_num: int, block_size: int, total_size: int) -> None:
        if total_size <= 0:
            return
        downloaded = min(block_num * block_size, total_size)
        percent = int((100 * downloaded) / total_size)

        now = time.monotonic()
        if last_bytes[0] is None or last_time[0] is None:
            speed_text = "?? MB/s"
        else:
            delta_bytes = max(downloaded - int(last_bytes[0]), 0)
            delta_time = max(now - float(last_time[0]), 1e-6)
            bytes_per_sec = int(delta_bytes / delta_time)
            if bytes_per_sec >= 1024 * 1024:
                speed_text = f"{bytes_per_sec // (1024 * 1024)} MB/s"
            else:
                speed_text = f"{bytes_per_sec // 1024} KB/s"

        last_bytes[0] = downloaded
        last_time[0] = now

        if percent != last_percent[0]:
            last_percent[0] = percent
            print(f"\rDownloading {path.name}: {percent}% ({speed_text})", end="", flush=True)

    urllib.request.urlretrieve(url, path, reporthook=_reporthook)
    if last_percent[0] >= 0:
        print(flush=True)
    return path


def resolve_cache_dir(cache_dir: str | Path | None, *, default_name: str) -> Path:
    """Return a writable cache directory, creating it if needed."""
    base = Path(cache_dir) if cache_dir is not None else Path.home() / ".cache" / default_name
    base.mkdir(parents=True, exist_ok=True)
    return base


def ensure_file(path: Path, url: str) -> Path:
    """Download ``url`` into ``path`` if it doesn't already exist."""
    if path.exists():
        return path
    return download_with_progress(url, path)


def ensure_csv(
    cache_dir: Path,
    filename: str,
    url: str,
    data_path: Optional[str | Path] = None,
) -> Path:
    """Resolve a CSV file via cache download or user-provided path."""
    if data_path is not None:
        return Path(data_path)
    target = cache_dir / filename
    if not target.exists():
        download_with_progress(url, target)
    return target


def download_archive(url: str, path: Path) -> Path:
    """Download an archive to ``path`` if needed."""
    return ensure_file(path, url)


def ensure_tar_extracted(archive: Path, extract_root: Path, target_dir: str) -> Path:
    """Extract a tar.gz archive into ``extract_root`` and return the target dir."""
    target_path = extract_root / target_dir
    if target_path.exists():
        return target_path
    extract_root.mkdir(parents=True, exist_ok=True)
    with tarfile.open(archive, "r:gz") as tar:
        tar.extractall(path=extract_root)
    return target_path


def to_host_jax_array(array: np.ndarray) -> jax.Array:
    """Copy a NumPy array onto the default CPU device for JAX consumption."""
    cpu_devices = jax.devices("cpu")
    if cpu_devices:
        with jax.default_device(cpu_devices[0]):
            return jnp.asarray(array)
    return jnp.asarray(array)



"""For those premature optimizers, you can `jit` your entire training epoch.

```python
import jax
import jax.numpy as jnp

from cyreal.sources import ArraySource
from cyreal.transforms import BatchTransform, DevicePutTransform
from cyreal.loader import DataLoader
from cyreal.datasets import MNISTDataset

train_data = MNISTDataset(split="test").as_array_dict()
pipeline = [
    ArraySource(train_data, ordering="shuffle"),
    BatchTransform(batch_size=128),
    DevicePutTransform(),
]
loader = DataLoader(pipeline)
loader_state = loader.init_state(jax.random.key(0))

model_state = {"params": jnp.array(0)}

@jax.jit
def train_epoch(model_state, loader_state):
    def model_update(model_state, batch, mask):
        # Update the network using your train fn
        new_model_state = {"params": model_state['params'] + 1}
        return new_model_state, None

    # scan_epoch is a helper method, but the loader itself is fully JIT-compatible
    # in case you want to roll your own training loop.
    loader_state, model_state, _ = loader.scan_epoch(loader_state, model_state, model_update)
    return model_state, loader_state

model_state, loader_state = train_epoch(model_state, loader_state)
```
"""




"""Write `torch`-style dataloaders without `torch`. Here is an in-memory MNIST example.

```python
import jax
import jax.numpy as jnp

from cyreal.sources import ArraySource
from cyreal.transforms import (
  BatchTransform,
  DevicePutTransform,
)
from cyreal.loader import DataLoader
from cyreal.datasets import MNISTDataset

# Start with a dataset
train_data = MNISTDataset(split="train").as_array_dict()
pipeline = [
  # Construct an in-memory array from the dataset
  ArraySource(train_data, ordering="shuffle"),
  # Batch the data
  BatchTransform(batch_size=128),
  # Move the batch onto the GPU
  DevicePutTransform(),
]
# Construct the dataloader
loader = DataLoader(pipeline=pipeline)
# Initialize the loader state
state = loader.init_state(jax.random.key(0))

# Iterate over the data for one epoch
for batch, mask in loader.iterate(state):
  ...  # train your network!
```
"""


"""High-level helpers for building jittable data pipelines."""
from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any, Callable, Tuple

import jax
import jax.numpy as jnp

from .sources import Source

PyTree = Any

@jax.tree_util.register_pytree_node_class
@dataclass
class _LoaderState:
    inner_state: Any

    def tree_flatten(self):
        return (self.inner_state,), None

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        (inner_state,) = children
        return cls(inner_state=inner_state)


class _LoaderIterator:
    """Python iterator that walks a loader state for a fixed number of steps."""

    def __init__(self, loader: DataLoader, state: _LoaderState, steps: int | None) -> None:
        self._loader = loader
        self._state = state
        self._remaining = steps

    def __iter__(self):
        return self

    def __next__(self):
        if self._remaining is not None:
            if self._remaining <= 0:
                raise StopIteration
            self._remaining -= 1

        batch, self._state, mask = self._loader.next(self._state)
        return batch, mask

    @property
    def state(self) -> _LoaderState:
        return self._state


@dataclass
class DataLoader:
    """Composable pipeline constructed from explicit stages."""

    pipeline: Source | Sequence[Any]
    """Either a Source or a sequence of stages to compose into a data pipeline."""

    def __post_init__(self) -> None:
        self._source = self._coerce_pipeline(self.pipeline)
        self.steps_per_epoch = self._source.steps_per_epoch

    def _coerce_pipeline(self, pipeline: Source | Sequence[Any]) -> Source:
        if self._looks_like_source(pipeline):
            return pipeline  # type: ignore[return-value]

        if isinstance(pipeline, Sequence):
            if not pipeline:
                raise ValueError("Pipeline sequence must contain at least one stage.")
            stages = list(pipeline)
            head = stages[0]
            if not self._looks_like_source(head):
                raise TypeError("First pipeline stage must implement the Source protocol.")
            current: Source = head  # type: ignore[assignment]
            for stage in stages[1:]:
                current = self._attach_stage(current, stage)
            return current

        raise TypeError("`pipeline` must be a Source or a sequence of stages.")

    @staticmethod
    def _looks_like_source(candidate: Any) -> bool:
        return all(hasattr(candidate, attr) for attr in ("init_state", "next", "steps_per_epoch"))

    def _attach_stage(self, current: Source, stage: Any) -> Source:
        radd = getattr(stage, "__radd__", None)
        if callable(radd):
            bound = radd(current)
            if bound is not NotImplemented:
                if not self._looks_like_source(bound):
                    raise TypeError("Stage did not return a valid Source when composed via '+'.")
                return bound

        if callable(stage):
            bound = stage(current)
            if not self._looks_like_source(bound):
                raise TypeError("Callable pipeline stage must return a Source.")
            return bound

        raise TypeError("Pipeline entries after the first must be transforms or callables returning Sources.")

    def init_state(self, key: jax.Array) -> _LoaderState:
        inner_state = self._source.init_state(key)
        return _LoaderState(inner_state=inner_state)

    def next(self, state: _LoaderState) -> Tuple[PyTree, _LoaderState, jax.Array]:
        batch, mask, inner_state = self._source.next(state.inner_state)
        return batch, _LoaderState(inner_state=inner_state), mask

    def iterate(self, state: _LoaderState, *, steps: int | None = None) -> _LoaderIterator:
        """Return a Python iterator over loader outputs.

        for batch, mask in loader.iterate(state):
            ...

        WARNING: This method is very slow compared to using `loader.scan_epoch` or jitting `loader.next`.

        Args:
            state: Starting loader state.
            steps: Number of steps (updates) to iterate; defaults to a single epoch.
                Pass ``None`` to iterate indefinitely.
        """

        if steps is None:
            steps = self.steps_per_epoch
        elif steps < 0:
            raise ValueError("steps must be non-negative or None.")
        return _LoaderIterator(self, state, steps)

    def scan_epoch(
        self,
        state: _LoaderState,
        carry: Any,
        body_fn: Callable[[Any, PyTree, jax.Array], Tuple[Any, Any]],
    ):
        """Run a full epoch via `jax.lax.scan` with constant-shape batches.
        
        def body_fn(model_state, batch, mask):
            ...
            return new_model_state, None

        state, batch = loader.scan_epoch(state, model_state, body_fn)
        """

        def _body(loop_state, _):
            loader_state, loop_carry = loop_state
            batch, loader_state, mask = self.next(loader_state)
            new_carry, output = body_fn(loop_carry, batch, mask)
            return (loader_state, new_carry), output

        (loader_state, final_carry), outputs = jax.lax.scan(
            _body, (state, carry), jnp.arange(self.steps_per_epoch)
        )
        return loader_state, final_carry, outputs


"""`HostCallbackTransform` allows you to log metrics and call other impure IO within jit.

```python
import jax
import jax.numpy as jnp
import numpy as np

from cyreal.sources import ArraySource
from cyreal.transforms import BatchTransform, HostCallbackTransform
from cyreal.loader import DataLoader
from cyreal.datasets import MNISTDataset

def model(images):
    return jnp.mean(images.astype(jnp.float32), axis=(1, 2, 3))

def cross_entropy(logits, labels):
    labels = labels.astype(jnp.float32)
    return (logits - labels) ** 2

def log_loss(batch, mask):
    logits = model(batch["image"])
    loss = jnp.mean(cross_entropy(logits, batch["label"]) * mask[:, None])
    print("loss:", float(np.asarray(loss)))
    return batch

loader = DataLoader(
    pipeline=[
        ArraySource(MNISTDataset(split="train").as_array_dict(), ordering="shuffle"),
        BatchTransform(batch_size=128),
        HostCallbackTransform(fn=log_loss),
    ],
)
# Still jittable
state = loader.init_state(jax.random.key(0))
sample, mask, state = jax.jit(loader.next)(state)
```
"""



"""Shared helpers for IDX-formatted vision datasets."""
from __future__ import annotations

import gzip
import shutil
import struct
from pathlib import Path

import numpy as np


def read_idx_images(path: Path) -> np.ndarray:
    with gzip.open(path, "rb") as f:
        header = f.read(16)
        magic, num, rows, cols = struct.unpack(">IIII", header)
        if magic != 2051:
            raise ValueError(f"Unexpected image file magic number: {magic}.")
        data = np.frombuffer(f.read(), dtype=np.uint8)
    return data.reshape(num, rows, cols)


def read_idx_labels(path: Path) -> np.ndarray:
    with gzip.open(path, "rb") as f:
        header = f.read(8)
        magic, num = struct.unpack(">II", header)
        if magic != 2049:
            raise ValueError(f"Unexpected label file magic number: {magic}.")
        data = np.frombuffer(f.read(), dtype=np.uint8)
    return data.reshape(num)


def ensure_uncompressed_idx(path: Path) -> Path:
    target = path.with_suffix("")
    if target.exists():
        return target
    with gzip.open(path, "rb") as src, open(target, "wb") as dst:
        shutil.copyfileobj(src, dst)
    return target


def read_image_header(path: Path) -> tuple[int, int, int]:
    with open(path, "rb") as f:
        header = f.read(16)
        magic, num, rows, cols = struct.unpack(">IIII", header)
    if magic != 2051:
        raise ValueError(f"Unexpected image file magic number: {magic}.")
    return int(num), int(rows), int(cols)


def read_label_header(path: Path) -> int:
    with open(path, "rb") as f:
        header = f.read(8)
        magic, num = struct.unpack(">II", header)
    if magic != 2049:
        raise ValueError(f"Unexpected label file magic number: {magic}.")
    return int(num)



"""Streaming dataset sources inspired by Grain's paradigm."""
from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Callable, Literal, Protocol, TypeVar

import jax
import jax.numpy as jnp
import numpy as np
from jax import tree_util
from jax.experimental import io_callback

PyTree = Any
StateT = TypeVar("StateT")


class Source(Protocol[StateT]):
    """Interface for stateful, JIT-friendly data streams.

    A source exposes fixed-length epochs (`steps_per_epoch`) and provides
    methods for initializing internal state and iteratively producing samples
    plus boolean masks that denote whether a sample is valid (e.g. under
    padding).
    """

    steps_per_epoch: int
    """Number of items emitted per epoch."""

    def init_state(self, key: jax.Array) -> StateT:
        """Return an initial state for the source.

        Args:
            key: Optional PRNG key used for randomized behavior such as
                shuffling. Implementations should fall back to a default key
                when ``None`` is provided.

        Returns:
            Backend-specific state object that must be passed to ``next``.
        """
        ...

    def next(self, state: StateT) -> tuple[PyTree, jax.Array, StateT]:
        """Advance the stream and return the next value.

        Args:
            state: Previously-initialized source state.

        Returns:
            Tuple ``(value, mask, new_state)`` where ``value`` is a PyTree of
            arrays, ``mask`` is a boolean array indicating whether the sample
            is valid, and ``new_state`` should be provided to the next call.
        """
        ...

    def element_spec(self) -> PyTree:
        """PyTree of :class:`jax.ShapeDtypeStruct` describing emitted samples."""
        ...


@jax.tree_util.register_pytree_node_class
@dataclass
class _ArraySourceState:
    indices: jax.Array
    mask: jax.Array
    position: jax.Array
    key: jax.Array
    epoch: jax.Array

    def tree_flatten(self):
        return (self.indices, self.mask, self.position, self.key, self.epoch), None

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        indices, mask, position, key, epoch = children
        return cls(indices=indices, mask=mask, position=position, key=key, epoch=epoch)


@jax.tree_util.register_pytree_node_class
@dataclass
class _DiskSourceState:
    indices: jax.Array
    position: jax.Array
    key: jax.Array
    epoch: jax.Array
    buffer: PyTree
    buffer_pos: jax.Array
    buffer_count: jax.Array

    def tree_flatten(self):
        buffer_leaves, buffer_def = tree_util.tree_flatten(self.buffer)
        children = (
            self.indices,
            self.position,
            self.key,
            self.epoch,
            *buffer_leaves,
            self.buffer_pos,
            self.buffer_count,
        )
        return children, buffer_def

    @classmethod
    def tree_unflatten(cls, buffer_def, children):
        indices, position, key, epoch, *rest = children
        buffer_leaf_count = buffer_def.num_leaves if buffer_def is not None else 0
        buffer_leaves = rest[:buffer_leaf_count]
        buffer = (
            tree_util.tree_unflatten(buffer_def, buffer_leaves)
            if buffer_def is not None
            else None
        )
        buffer_pos, buffer_count = rest[buffer_leaf_count:]
        return cls(
            indices=indices,
            position=position,
            key=key,
            epoch=epoch,
            buffer=buffer,
            buffer_pos=buffer_pos,
            buffer_count=buffer_count,
        )


@dataclass
class ArraySource(Source[_ArraySourceState]):
    """Sample-level stream over an in-memory PyTree of arrays.

    Args:
        data: PyTree whose leaves are arrays with a leading sample dimension.
        ordering: Either ``"sequential"`` or ``"shuffle"``.
    """

    data: PyTree
    ordering: Literal["sequential", "shuffle"] = "shuffle"

    def __post_init__(self) -> None:
        leaves, self._treedef = tree_util.tree_flatten(self.data)
        if not leaves:
            raise ValueError("Data tree must contain at least one array.")

        first = leaves[0]
        self._num_samples = int(first.shape[0])
        for leaf in leaves[1:]:
            if leaf.shape[0] != self._num_samples:
                raise ValueError("All leaves must share the leading dimension.")
        if self._num_samples == 0:
            raise ValueError("Dataset cannot be empty.")

        self.steps_per_epoch = self._num_samples
        self._mask_template = jnp.ones(self._num_samples, dtype=bool)
        self._element_spec = tree_util.tree_map(
            lambda leaf: jax.ShapeDtypeStruct(shape=leaf.shape[1:], dtype=leaf.dtype),
            self.data,
        )

    @property
    def num_samples(self) -> int:
        return self._num_samples

    def element_spec(self) -> PyTree:
        """Shape/dtype metadata describing samples produced by the source."""
        return self._element_spec

    def _build_epoch_indices(self, key: jax.Array) -> tuple[jax.Array, jax.Array]:
        base = jnp.arange(self._num_samples)
        if self.ordering == "shuffle":
            base = jax.random.permutation(key, base)
        elif self.ordering != "sequential":
            raise ValueError(f"Unknown ordering '{self.ordering}'.")

        return base, self._mask_template

    def init_state(self, key: jax.Array) -> _ArraySourceState:
        """Create the initial iteration state.

        Args:
            key: Optional PRNG key. Defaults to ``jax.random.PRNGKey(0)`` when
                omitted.
        """
        key, perm_key = jax.random.split(key)
        indices, mask = self._build_epoch_indices(perm_key)
        position = jnp.array(0, dtype=jnp.int32)
        epoch = jnp.array(0, dtype=jnp.int32)
        return _ArraySourceState(indices=indices, mask=mask, position=position, key=key, epoch=epoch)

    def next(self, state: _ArraySourceState) -> tuple[PyTree, jax.Array, _ArraySourceState]:
        """Return the next sample (with mask) and the advanced state."""
        index = jax.lax.dynamic_index_in_dim(state.indices, state.position, axis=0, keepdims=False)
        mask_value = jax.lax.dynamic_index_in_dim(state.mask, state.position, axis=0, keepdims=False)
        sample = tree_util.tree_map(
            lambda arr: jax.lax.dynamic_index_in_dim(arr, index, axis=0, keepdims=False),
            self.data,
        )

        next_position = state.position + 1

        def _reset_epoch(_: None):
            new_key, perm_key = jax.random.split(state.key)
            indices, mask = self._build_epoch_indices(perm_key)
            return _ArraySourceState(
                indices=indices,
                mask=mask,
                position=jnp.array(0, dtype=jnp.int32),
                key=new_key,
                epoch=state.epoch + 1,
            )

        def _advance(_: None):
            return _ArraySourceState(
                indices=state.indices,
                mask=state.mask,
                position=next_position,
                key=state.key,
                epoch=state.epoch,
            )

        need_reset = next_position >= state.indices.shape[0]
        new_state = jax.lax.cond(need_reset, _reset_epoch, _advance, operand=None)
        return sample, mask_value, new_state


@dataclass
class DiskSource(Source[_DiskSourceState]):
    """Sample-level stream that loads items via a Python callback (disk, RPC, etc.).

    This is slow, only use this if your dataset will not fit in system memory.
    """

    length: int
    """Number of samples in the dataset."""
    sample_fn: Callable[[int], PyTree]
    """Python callable that takes an integer index and returns a PyTree of arrays."""
    sample_spec: PyTree | None = None
    """Optional PyTree of `jax.ShapeDtypeStruct` describing the shape and dtype of samples."""
    ordering: Literal["sequential", "shuffle"] = "shuffle"
    """Sample ordering strategy, either 'sequential' or 'shuffle'. The shuffling occurs over the entire dataset, not within the prefetch buffer."""
    prefetch_size: int = 64
    """Number of samples to prefetch into a JAX array buffer. Set this larger to achieve better throughput at the cost of more memory usage."""

    def __post_init__(self) -> None:
        if self.length <= 0:
            raise ValueError("Dataset cannot be empty.")
        if self.prefetch_size <= 0:
            raise ValueError("prefetch_size must be positive.")

        if self.sample_spec is None:
            example = self.sample_fn(0)

            def _to_spec(leaf):
                arr = np.asarray(leaf)
                return jax.ShapeDtypeStruct(shape=arr.shape, dtype=arr.dtype)

            self.sample_spec = tree_util.tree_map(_to_spec, example)

        leaves = tree_util.tree_leaves(self.sample_spec)
        if not leaves:
            raise ValueError("element_spec must include at least one leaf.")
        for leaf in leaves:
            if not isinstance(leaf, jax.ShapeDtypeStruct):
                raise TypeError("element_spec leaves must be jax.ShapeDtypeStruct instances.")

        self._num_samples = int(self.length)
        self.steps_per_epoch = self._num_samples
        self._element_spec = self.sample_spec
        self.prefetch_size = int(self.prefetch_size)

        def _zeros(spec: jax.ShapeDtypeStruct):
            return np.zeros(spec.shape, dtype=np.dtype(spec.dtype))

        def _buffer_shape(spec: jax.ShapeDtypeStruct):
            return jax.ShapeDtypeStruct(
                shape=(self.prefetch_size, *spec.shape),
                dtype=spec.dtype,
            )

        self._zero_sample = tree_util.tree_map(_zeros, self.sample_spec)
        self._chunk_spec = tree_util.tree_map(_buffer_shape, self.sample_spec)
        self._buffer_template = tree_util.tree_map(
            lambda spec: jnp.zeros((self.prefetch_size, *spec.shape), dtype=spec.dtype),
            self.sample_spec,
        )

    def element_spec(self) -> PyTree:
        """Shape/dtype metadata describing samples produced by the source."""
        return self._element_spec

    def _build_epoch_indices(self, key: jax.Array) -> jax.Array:
        base = jnp.arange(self._num_samples)
        if self.ordering == "shuffle":
            base = jax.random.permutation(key, base)
        elif self.ordering != "sequential":
            raise ValueError(f"Unknown ordering '{self.ordering}'.")
        return base

    def init_state(self, key: jax.Array) -> _DiskSourceState:
        """Build the starting state, optionally seeding randomness with ``key``."""
        key, perm_key = jax.random.split(key)
        indices = self._build_epoch_indices(perm_key)
        position = jnp.array(0, dtype=jnp.int32)
        epoch = jnp.array(0, dtype=jnp.int32)
        return _DiskSourceState(
            indices=indices,
            position=position,
            key=key,
            epoch=epoch,
            buffer=self._buffer_template,
            buffer_pos=jnp.array(0, dtype=jnp.int32),
            buffer_count=jnp.array(0, dtype=jnp.int32),
        )

    def _chunk_callback(self, indices: np.ndarray, mask: np.ndarray) -> PyTree:
        idx_array = np.asarray(indices, dtype=np.int64)
        mask_array = np.asarray(mask, dtype=bool)
        samples: list[PyTree] = []
        for keep, idx in zip(mask_array, idx_array):
            if keep:
                samples.append(self.sample_fn(int(idx)))
            else:
                samples.append(self._zero_sample)
        return tree_util.tree_map(lambda *xs: np.stack(xs, axis=0), *samples)

    def _maybe_reset_epoch(self, state: _DiskSourceState) -> _DiskSourceState:
        def _reset(state: _DiskSourceState):
            new_key, perm_key = jax.random.split(state.key)
            indices = self._build_epoch_indices(perm_key)
            return _DiskSourceState(
                indices=indices,
                position=jnp.array(0, dtype=jnp.int32),
                key=new_key,
                epoch=state.epoch + 1,
                buffer=self._buffer_template,
                buffer_pos=jnp.array(0, dtype=jnp.int32),
                buffer_count=jnp.array(0, dtype=jnp.int32),
            )

        return jax.lax.cond(state.position >= self._num_samples, _reset, lambda s: s, state)

    def _maybe_refill_buffer(self, state: _DiskSourceState) -> _DiskSourceState:
        def _needs(state: _DiskSourceState):
            return jnp.logical_or(state.buffer_count == 0, state.buffer_pos >= state.buffer_count)

        def _refill(state: _DiskSourceState):
            refreshed = self._maybe_reset_epoch(state)
            remaining = self._num_samples - refreshed.position
            chunk = jnp.minimum(remaining, self.prefetch_size)
            chunk = jnp.maximum(chunk, 0)
            chunk = chunk.astype(jnp.int32)
            offsets = jnp.arange(self.prefetch_size, dtype=jnp.int32)
            gather_positions = jnp.minimum(
                refreshed.position + offsets,
                refreshed.indices.shape[0] - 1,
            )
            chunk_indices = jax.vmap(
                lambda idx: jax.lax.dynamic_index_in_dim(
                    refreshed.indices, idx, axis=0, keepdims=False
                )
            )(gather_positions)
            valid_mask = offsets < chunk
            buffer = io_callback(self._chunk_callback, self._chunk_spec, chunk_indices, valid_mask)
            new_position = refreshed.position + chunk
            return _DiskSourceState(
                indices=refreshed.indices,
                position=new_position,
                key=refreshed.key,
                epoch=refreshed.epoch,
                buffer=buffer,
                buffer_pos=jnp.array(0, dtype=jnp.int32),
                buffer_count=chunk,
            )

        return jax.lax.cond(_needs(state), _refill, lambda s: s, state)

    def next(self, state: _DiskSourceState) -> tuple[PyTree, jax.Array, _DiskSourceState]:
        """Return buffered sample, all-True mask, and updated state."""
        state = self._maybe_refill_buffer(state)
        sample = tree_util.tree_map(
            lambda buf: jax.lax.dynamic_index_in_dim(
                buf, state.buffer_pos, axis=0, keepdims=False
            ),
            state.buffer,
        )
        mask_value = jnp.array(True, dtype=bool)
        new_state = _DiskSourceState(
            indices=state.indices,
            position=state.position,
            key=state.key,
            epoch=state.epoch,
            buffer=state.buffer,
            buffer_pos=state.buffer_pos + 1,
            buffer_count=state.buffer_count,
        )
        return sample, mask_value, new_state




@jax.tree_util.register_pytree_node_class
@dataclass
class GymnaxSourceState:
    env_state: PyTree
    obs: PyTree
    key: jax.Array
    step: jax.Array
    epoch: jax.Array
    policy_state: PyTree | None = None
    new_episode: jax.Array | None = None

    def tree_flatten(self):
        return (
            self.env_state,
            self.obs,
            self.key,
            self.step,
            self.epoch,
            self.policy_state,
            self.new_episode,
        ), None

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        env_state, obs, key, step, epoch, policy_state, new_episode = children
        return cls(
            env_state=env_state,
            obs=obs,
            key=key,
            step=step,
            epoch=epoch,
            policy_state=policy_state,
            new_episode=new_episode,
        )


@dataclass
class GymnaxSource(Source[GymnaxSourceState]):
    """Stream transitions by rolling out a Gymnax environment with a policy.
    
    Useful for reinforcement learning.

    Args:
        env: Gymnax environment instance.
        env_params: Parameters to pass to the environment's reset and step functions.
        policy_step_fn: Callable that takes (observation, policy_state, new_episode, key) and
            returns (action, new_policy_state).
        policy_state_template: Example PyTree carrying everything required by
            ``policy_step_fn`` (for example, policy parameters and recurrent
            carries). This template is used only to infer the element spec; callers
            are responsible for injecting a real policy state into the loader
            state before calling ``next``.
        steps_per_epoch: Number of environment steps per epoch for a single environment.
    """

    env: Any
    env_params: Any
    policy_step_fn: Callable[[PyTree, PyTree, jax.Array, jax.Array], tuple[PyTree, PyTree]]
    policy_state_template: PyTree | None = None
    steps_per_epoch: int = 1024

    def __post_init__(self) -> None:
        if self.steps_per_epoch <= 0:
            raise ValueError("steps_per_epoch must be positive.")
        if self.policy_state_template is None:
            raise ValueError("GymnaxSource requires a policy_state_template for shape inference.")

        def _sample(key, policy_state):
            obs, env_state = self.env.reset(key, self.env_params)
            action, next_policy_state = self.policy_step_fn(
                obs,
                policy_state,
                jnp.array(True, dtype=jnp.bool_),
                key,
            )
            next_obs, _, reward, done, info = self.env.step(
                key,
                env_state,
                action,
                self.env_params,
            )
            transition = {
                "state": obs,
                "action": action,
                "reward": reward,
                "next_state": next_obs,
                "done": done,
                "info": info,
            }
            return transition, next_policy_state

        shaped, _ = jax.eval_shape(_sample, jax.random.PRNGKey(0), self.policy_state_template)
        self._element_spec = tree_util.tree_map(
            lambda arr: jax.ShapeDtypeStruct(shape=arr.shape, dtype=arr.dtype), shaped
        )
        self.policy_state_template = None

    def element_spec(self) -> PyTree:
        """Shape/dtype metadata describing Gymnax transitions."""
        return self._element_spec

    def init_state(self, key: jax.Array) -> GymnaxSourceState:
        """Return RNG-seeded environment + policy state for iteration."""
        key, env_key = jax.random.split(key)
        obs, env_state = self.env.reset(env_key, self.env_params)
        return GymnaxSourceState(
            env_state=env_state,
            obs=obs,
            key=key,
            step=jnp.array(0, dtype=jnp.int32),
            epoch=jnp.array(0, dtype=jnp.int32),
            policy_state=None,
            new_episode=jnp.array(True, dtype=jnp.bool_),
        )

    def next(self, state: GymnaxSourceState) -> tuple[PyTree, jax.Array, GymnaxSourceState]:
        """Roll the environment forward one step and emit a transition."""
        key, policy_key, step_key, done_reset_key, epoch_reset_key = jax.random.split(state.key, 5)

        if state.policy_state is None:
            raise ValueError(
                "GymnaxSource state is missing `policy_state`; set it explicitly before calling `next`."
            )
        policy_state = state.policy_state

        if state.new_episode is None:
            raise ValueError("GymnaxSource state is missing `new_episode` flag.")

        action, updated_policy_state = self.policy_step_fn(
            state.obs,
            policy_state,
            state.new_episode,
            policy_key,
        )
        next_obs, next_env_state, reward, done, info = self.env.step(
            step_key,
            state.env_state,
            action,
            self.env_params,
        )

        transition = {
            "state": state.obs,
            "action": action,
            "reward": reward,
            "next_state": next_obs,
            "done": done,
            "info": info,
        }
        mask = jnp.array(True, dtype=bool)

        done_flag = jnp.asarray(done, dtype=bool)
        done_flag = jnp.reshape(done_flag, ())
        reset_obs, reset_env_state = self.env.reset(done_reset_key, self.env_params)

        cont_obs, cont_env_state = jax.lax.cond(
            done_flag,
            lambda _: (reset_obs, reset_env_state),
            lambda _: (next_obs, next_env_state),
            operand=None,
        )

        next_step = state.step + 1
        need_epoch_reset = next_step >= self.steps_per_epoch

        def _reset_epoch(_: None):
            epoch_obs, epoch_env_state = self.env.reset(epoch_reset_key, self.env_params)
            return GymnaxSourceState(
                env_state=epoch_env_state,
                obs=epoch_obs,
                key=key,
                step=jnp.array(0, dtype=jnp.int32),
                epoch=state.epoch + 1,
                policy_state=updated_policy_state,
                new_episode=jnp.array(True, dtype=jnp.bool_),
            )

        def _continue(_: None):
            return GymnaxSourceState(
                env_state=cont_env_state,
                obs=cont_obs,
                key=key,
                step=next_step,
                epoch=state.epoch,
                policy_state=updated_policy_state,
                new_episode=done_flag,
            )

        new_state = jax.lax.cond(need_epoch_reset, _reset_epoch, _continue, operand=None)
        return transition, mask, new_state


"""Streaming from disk using `DiskSource`. This is much slower than an in-memory dataset
but it allows you to work with datasets that do not fit into RAM. The key is to call
`make_disk_source` on your dataset class to get a disk-backed source.

```python
import jax

from cyreal.transforms import BatchTransform, DevicePutTransform
from cyreal.loader import DataLoader
from cyreal.datasets import MNISTDataset

pipeline = [
    # Prefetch 1024 examples for each disk read
    MNISTDataset.make_disk_source(split="train", ordering="shuffle", prefetch_size=1024),
    BatchTransform(batch_size=128),
    DevicePutTransform(),
]

loader = DataLoader(pipeline=pipeline)
state = loader.init_state(jax.random.PRNGKey(0))

for batch, mask in loader.iterate(state):
    ...  # stream without holding the dataset in RAM
```
"""



"""Tests covering dataset utilities and disk sources."""
from __future__ import annotations

import gzip
import pickle
import struct
from pathlib import Path

import jax
import jax.numpy as jnp
import numpy as np
import pytest

from cyreal.transforms import BatchTransform
from cyreal.datasets import (
    CelebADataset,
    CIFAR10Dataset,
    CIFAR100Dataset,
    EMNISTDataset,
    FashionMNISTDataset,
    KMNISTDataset,
    MNISTDataset,
)


def _write_idx_images(path, images):
    with gzip.open(path, "wb") as f:
        num, rows, cols = images.shape
        f.write(struct.pack(">IIII", 2051, num, rows, cols))
        f.write(images.tobytes())


def _write_idx_labels(path, labels):
    with gzip.open(path, "wb") as f:
        num = labels.shape[0]
        f.write(struct.pack(">II", 2049, num))
        f.write(labels.tobytes())


def _seed_fake_cifar10(tmp_path, split="train", samples_per_batch=1):
    archive_path = tmp_path / "cifar-10-python.tar.gz"
    archive_path.write_bytes(b"")
    batches_dir = tmp_path / "cifar-10-batches-py"
    batches_dir.mkdir(parents=True, exist_ok=True)

    if split == "train":
        names = [f"data_batch_{i}" for i in range(1, 6)]
    elif split == "test":
        names = ["test_batch"]
    else:
        raise ValueError("split must be 'train' or 'test'.")

    current_label = 0
    for name in names:
        data = []
        labels = []
        for _ in range(samples_per_batch):
            pixel_value = current_label % 256
            image = np.full((3, 32, 32), pixel_value, dtype=np.uint8)
            data.append(image.reshape(-1))
            labels.append(current_label)
            current_label += 1
        batch = {"data": np.stack(data, axis=0), "labels": labels}
        with open(batches_dir / name, "wb") as f:
            pickle.dump(batch, f, protocol=2)


def _seed_fake_cifar100(tmp_path, split="train", samples=4):
    archive_path = tmp_path / "cifar-100-python.tar.gz"
    archive_path.write_bytes(b"")
    target_dir = tmp_path / "cifar-100-python"
    target_dir.mkdir(parents=True, exist_ok=True)

    if split not in {"train", "test"}:
        raise ValueError("split must be 'train' or 'test'.")

    data = []
    fine_labels = []
    coarse_labels = []
    for idx in range(samples):
        pixel_value = idx % 256
        image = np.full((3, 32, 32), pixel_value, dtype=np.uint8)
        data.append(image.reshape(-1))
        fine_labels.append(idx)
        coarse_labels.append(idx % 5)
    batch = {
        "data": np.stack(data, axis=0),
        "fine_labels": fine_labels,
        "coarse_labels": coarse_labels,
    }
    with open(target_dir / split, "wb") as f:
        pickle.dump(batch, f, protocol=2)


def _seed_fake_celeba(tmp_path: Path):
    pil = pytest.importorskip("PIL.Image")
    image_module = pil

    image_dir = tmp_path / "img_align_celeba"
    image_dir.mkdir(parents=True, exist_ok=True)

    samples = [
        ("000001.jpg", 0, np.array([1, -1, 1], dtype=np.int8), 10),
        ("000002.jpg", 1, np.array([-1, 1, -1], dtype=np.int8), 40),
        ("000003.jpg", 2, np.array([1, 1, -1], dtype=np.int8), 90),
    ]

    for name, _, _, pixel in samples:
        image = np.full((4, 5, 3), pixel, dtype=np.uint8)
        image_module.fromarray(image, mode="RGB").save(image_dir / name)

    with (tmp_path / "list_eval_partition.txt").open("w", encoding="utf-8") as f:
        f.write(f"{len(samples)}\n")
        for name, partition, _, _ in samples:
            f.write(f"{name} {partition}\n")

    with (tmp_path / "list_attr_celeba.txt").open("w", encoding="utf-8") as f:
        f.write(f"{len(samples)}\n")
        f.write("Smiling Young Male\n")
        for name, _, attrs, _ in samples:
            values = " ".join(str(int(v)) for v in attrs)
            f.write(f"{name} {values}\n")


MNIST_LIKE_DATASETS = [
    pytest.param(MNISTDataset, {}, id="mnist"),
    pytest.param(FashionMNISTDataset, {}, id="fashion"),
    pytest.param(KMNISTDataset, {}, id="kmnist"),
    pytest.param(EMNISTDataset, {"subset": "letters"}, id="emnist-letters"),
]


@pytest.mark.parametrize("dataset_cls,extra_kwargs", MNIST_LIKE_DATASETS)
def test_idx_dataset_reads_cached_idx(tmp_path, dataset_cls, extra_kwargs):
    num, rows, cols = 3, 2, 2
    images = np.arange(num * rows * cols, dtype=np.uint8).reshape(num, rows, cols)
    labels = np.arange(num, dtype=np.uint8)

    images_path = tmp_path / "train_images.gz"
    labels_path = tmp_path / "train_labels.gz"
    _write_idx_images(images_path, images)
    _write_idx_labels(labels_path, labels)

    dataset = dataset_cls(split="train", cache_dir=tmp_path, **extra_kwargs)
    assert len(dataset) == num
    example = dataset[0]
    img = example["image"]
    label = example["label"]
    assert img.shape == (rows, cols, 1)
    assert img.dtype == np.uint8
    assert label == 0
    np.testing.assert_array_equal(img[..., 0], images[0])


@pytest.mark.parametrize("dataset_cls,extra_kwargs", MNIST_LIKE_DATASETS)
def test_idx_disk_sources_stream_from_disk(tmp_path, dataset_cls, extra_kwargs):
    num, rows, cols = 3, 2, 2
    images = np.arange(num * rows * cols, dtype=np.uint8).reshape(num, rows, cols)
    labels = np.arange(num, dtype=np.uint8)

    images_path = tmp_path / "train_images.gz"
    labels_path = tmp_path / "train_labels.gz"
    _write_idx_images(images_path, images)
    _write_idx_labels(labels_path, labels)

    common_kwargs = {
        "split": "train",
        "cache_dir": tmp_path,
        "ordering": "sequential",
        "prefetch_size": 2,
    }
    common_kwargs.update(extra_kwargs)
    source = dataset_cls.make_disk_source(**common_kwargs)

    batched = BatchTransform(
        batch_size=2,
        element_spec_override=source.element_spec(),
    )(source)

    state = batched.init_state(jax.random.PRNGKey(0))
    batch, mask, state = batched.next(state)
    np.testing.assert_array_equal(np.asarray(batch["image"][..., 0]), images[:2])
    np.testing.assert_array_equal(np.asarray(batch["label"]), np.array([0, 1], dtype=np.int32))
    np.testing.assert_array_equal(np.asarray(mask), np.array([True, True]))

    batch2, mask2, _ = batched.next(state)
    np.testing.assert_array_equal(np.asarray(batch2["image"][0, ..., 0]), images[2])
    np.testing.assert_array_equal(np.asarray(batch2["label"][:1]), np.array([2], dtype=np.int32))
    np.testing.assert_array_equal(np.asarray(mask2), np.array([True, False]))


def test_cifar10_disk_source_streams_from_disk(tmp_path):
    _seed_fake_cifar10(tmp_path, split="test", samples_per_batch=3)

    source = CIFAR10Dataset.make_disk_source(
        split="test",
        cache_dir=tmp_path,
        ordering="sequential",
        prefetch_size=2,
    )
    batched = BatchTransform(
        batch_size=2,
        element_spec_override=source.element_spec(),
    )(source)

    state = batched.init_state(jax.random.PRNGKey(0))
    batch, mask, state = batched.next(state)
    np.testing.assert_array_equal(np.asarray(batch["label"]), np.array([0, 1], dtype=np.int32))
    np.testing.assert_array_equal(
        np.asarray(batch["image"][0, 0, 0]),
        np.array([0, 0, 0], dtype=np.uint8),
    )
    np.testing.assert_array_equal(
        np.asarray(batch["image"][1, 0, 0]),
        np.array([1, 1, 1], dtype=np.uint8),
    )
    np.testing.assert_array_equal(np.asarray(mask), np.array([True, True]))

    batch2, mask2, _ = batched.next(state)
    np.testing.assert_array_equal(np.asarray(batch2["label"][:1]), np.array([2], dtype=np.int32))
    np.testing.assert_array_equal(np.asarray(mask2), np.array([True, False]))
    np.testing.assert_array_equal(
        np.asarray(batch2["image"][1]),
        np.zeros((32, 32, 3), dtype=np.uint8),
    )


def test_cifar100_disk_source_streams_from_disk(tmp_path):
    _seed_fake_cifar100(tmp_path, split="test", samples=3)

    source = CIFAR100Dataset.make_disk_source(
        split="test",
        cache_dir=tmp_path,
        ordering="sequential",
        prefetch_size=2,
    )

    batched = BatchTransform(
        batch_size=2,
        element_spec_override=source.element_spec(),
    )(source)

    state = batched.init_state(jax.random.PRNGKey(0))
    batch, mask, state = batched.next(state)
    np.testing.assert_array_equal(np.asarray(batch["label"]), np.array([0, 1], dtype=np.int32))
    np.testing.assert_array_equal(np.asarray(batch["coarse_label"]), np.array([0, 1], dtype=np.int32))
    np.testing.assert_array_equal(
        np.asarray(batch["image"][0, 0, 0]),
        np.array([0, 0, 0], dtype=np.uint8),
    )
    np.testing.assert_array_equal(
        np.asarray(batch["image"][1, 0, 0]),
        np.array([1, 1, 1], dtype=np.uint8),
    )
    np.testing.assert_array_equal(np.asarray(mask), np.array([True, True]))

    batch2, mask2, _ = batched.next(state)
    np.testing.assert_array_equal(np.asarray(batch2["label"][:1]), np.array([2], dtype=np.int32))
    np.testing.assert_array_equal(np.asarray(batch2["coarse_label"][:1]), np.array([2], dtype=np.int32))
    np.testing.assert_array_equal(np.asarray(mask2), np.array([True, False]))


def test_celeba_dataset_reads_local_layout(tmp_path):
    _seed_fake_celeba(tmp_path)

    dataset = CelebADataset(split="valid", data_dir=tmp_path)
    assert len(dataset) == 1

    sample = dataset[0]
    assert sample["image"].shape == (4, 5, 3)
    assert sample["image"].dtype == np.uint8
    np.testing.assert_array_equal(
        np.asarray(sample["attributes"]),
        np.array([0, 1, 0], dtype=np.int32),
    )
    assert dataset.attribute_names == ("Smiling", "Young", "Male")


def test_celeba_disk_source_streams_from_disk(tmp_path):
    _seed_fake_celeba(tmp_path)

    source = CelebADataset.make_disk_source(
        split="test",
        data_dir=tmp_path,
        ordering="sequential",
        prefetch_size=2,
    )
    batched = BatchTransform(
        batch_size=2,
        element_spec_override=source.element_spec(),
    )(source)

    state = batched.init_state(jax.random.PRNGKey(0))
    batch, mask, _ = batched.next(state)
    np.testing.assert_array_equal(np.asarray(mask), np.array([True, False]))
    np.testing.assert_array_equal(
        np.asarray(batch["attributes"][0]),
        np.array([1, 1, 0], dtype=np.int32),
    )
    assert np.asarray(batch["image"])[0].shape == (4, 5, 3)


"""Sunspots time-series dataset built from CSV windows."""
from __future__ import annotations

from dataclasses import dataclass
from typing import Literal

import jax

from .dataset_protocol import DatasetProtocol
from ..sources import DiskSource
from .time_utils import make_sequence_disk_source, prepare_time_windows
from .utils import to_host_jax_array as _to_host_jax_array

SUNSPOTS_URL = (
    "https://raw.githubusercontent.com/jbrownlee/Datasets/master/monthly-sunspots.csv"
)


@dataclass
class SunspotsDataset(DatasetProtocol):
    """A time series classification dataset of sunspot counts from SILSO."""

    split: Literal["train", "test"] = "train"
    context_length: int = 24
    prediction_length: int = 1
    train_fraction: float = 0.8
    cache_dir: str | None = None
    data_path: str | None = None

    def __post_init__(self) -> None:
        contexts, targets = prepare_time_windows(
            dataset_name="sunspots",
            filename="monthly-sunspots.csv",
            url=SUNSPOTS_URL,
            skip_header=1,
            value_column=1,
            split=self.split,
            context_length=self.context_length,
            prediction_length=self.prediction_length,
            train_fraction=self.train_fraction,
            cache_dir=self.cache_dir,
            data_path=self.data_path,
        )
        self._contexts = _to_host_jax_array(contexts)
        self._targets = _to_host_jax_array(targets)

    def __len__(self) -> int:
        return int(self._contexts.shape[0])

    def __getitem__(self, index: int):
        return {
            "context": self._contexts[index],
            "target": self._targets[index],
        }

    def as_array_dict(self) -> dict[str, jax.Array]:
        """Expose the full dataset as a PyTree of JAX arrays."""
        return {"context": self._contexts, "target": self._targets}

    @classmethod
    def make_disk_source(
        cls,
        *,
        split: Literal["train", "test"] = "train",
        context_length: int = 24,
        prediction_length: int = 1,
        train_fraction: float = 0.8,
        cache_dir: str | None = None,
        data_path: str | None = None,
        ordering: Literal["sequential", "shuffle"] = "shuffle",
        prefetch_size: int = 64,
    ) -> DiskSource:
        """Return the dataset in a disk streaming format."""

        contexts, targets = prepare_time_windows(
            dataset_name="sunspots",
            filename="monthly-sunspots.csv",
            url=SUNSPOTS_URL,
            skip_header=1,
            value_column=1,
            split=split,
            context_length=context_length,
            prediction_length=prediction_length,
            train_fraction=train_fraction,
            cache_dir=cache_dir,
            data_path=data_path,
        )
        return make_sequence_disk_source(
            contexts=contexts,
            targets=targets,
            context_length=context_length,
            prediction_length=prediction_length,
            ordering=ordering,
            prefetch_size=prefetch_size,
        )



"""CIFAR-100 dataset helpers."""
from __future__ import annotations

import pickle
from dataclasses import dataclass
from pathlib import Path
from typing import Literal

import jax
import jax.numpy as jnp
import numpy as np

from .dataset_protocol import DatasetProtocol
from ..sources import DiskSource
from .utils import (
    download_archive,
    ensure_tar_extracted,
    resolve_cache_dir,
    to_host_jax_array as _to_host_jax_array,
)

CIFAR100_URL = "https://www.cs.toronto.edu/~kriz/cifar-100-python.tar.gz"


def _load_split_numpy(split: Literal["train", "test"], extract_dir: Path) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    file_name = "train" if split == "train" else "test"
    batch_path = extract_dir / file_name
    if not batch_path.exists():
        raise FileNotFoundError(f"Missing CIFAR-100 split file '{file_name}'.")
    with open(batch_path, "rb") as f:
        batch = pickle.load(f, encoding="latin1")
    data = batch["data"].reshape(-1, 3, 32, 32)
    images = np.transpose(data, (0, 2, 3, 1)).astype(np.uint8)
    fine = np.asarray(batch["fine_labels"], dtype=np.int32)
    coarse = np.asarray(batch["coarse_labels"], dtype=np.int32)
    return images, fine, coarse

def _ensure_split_numpy_cache(
    split: Literal["train", "test"],
    base_dir: Path,
    extract_dir: Path,
) -> tuple[Path, Path, Path]:
    cache_root = base_dir / "disk_cache"
    cache_root.mkdir(parents=True, exist_ok=True)
    images_path = cache_root / f"{split}_images.npy"
    fine_labels_path = cache_root / f"{split}_fine_labels.npy"
    coarse_labels_path = cache_root / f"{split}_coarse_labels.npy"
    if not images_path.exists() or not fine_labels_path.exists() or not coarse_labels_path.exists():
        images_np, fine_np, coarse_np = _load_split_numpy(split, extract_dir)
        np.save(images_path, images_np)
        np.save(fine_labels_path, fine_np)
        np.save(coarse_labels_path, coarse_np)
    return images_path, fine_labels_path, coarse_labels_path


@dataclass
class CIFAR100Dataset(DatasetProtocol):
    """The CIFAR-100 image classification dataset."""

    split: Literal["train", "test"] = "train"
    cache_dir: str | Path | None = None

    def __post_init__(self) -> None:
        base_dir = resolve_cache_dir(self.cache_dir, default_name="cifar100")
        archive_path = base_dir / "cifar-100-python.tar.gz"
        download_archive(CIFAR100_URL, archive_path)
        extract_dir = ensure_tar_extracted(archive_path, base_dir, target_dir="cifar-100-python")

        images_np, fine_np, coarse_np = _load_split_numpy(self.split, extract_dir)
        self._images = _to_host_jax_array(images_np)
        self._fine_labels = _to_host_jax_array(fine_np)
        self._coarse_labels = _to_host_jax_array(coarse_np)

    def __len__(self) -> int:
        return int(self._images.shape[0])

    def __getitem__(self, index: int):
        return {
            "image": self._images[index],
            "label": self._fine_labels[index],
            "coarse_label": self._coarse_labels[index],
        }

    def as_array_dict(self) -> dict[str, jax.Array]:
        """Expose the full dataset as a PyTree of JAX arrays."""
        return {
            "image": self._images,
            "label": self._fine_labels,
            "coarse_label": self._coarse_labels,
        }

    @classmethod
    def make_disk_source(
        cls,
        *,
        split: Literal["train", "test"] = "train",
        cache_dir: str | Path | None = None,
        ordering: Literal["sequential", "shuffle"] = "shuffle",
        prefetch_size: int = 64,
    ) -> DiskSource:
        """Return the dataset in a disk streaming format."""

        base_dir = resolve_cache_dir(cache_dir, default_name="cifar100")
        archive_path = base_dir / "cifar-100-python.tar.gz"
        download_archive(CIFAR100_URL, archive_path)
        extract_dir = ensure_tar_extracted(archive_path, base_dir, target_dir="cifar-100-python")

        images_path, fine_labels_path, coarse_labels_path = _ensure_split_numpy_cache(
            split, base_dir, extract_dir
        )
        images_memmap = np.load(images_path, mmap_mode="r")
        fine_labels_memmap = np.load(fine_labels_path, mmap_mode="r")
        coarse_labels_memmap = np.load(coarse_labels_path, mmap_mode="r")

        lengths = {
            images_memmap.shape[0],
            fine_labels_memmap.shape[0],
            coarse_labels_memmap.shape[0],
        }
        if len(lengths) != 1:
            raise ValueError("CIFAR-100 arrays do not share the same length.")

        def _read_sample(index: int | np.ndarray) -> dict[str, np.ndarray]:
            idx = int(np.asarray(index))
            image = np.asarray(images_memmap[idx], dtype=np.uint8)
            label = np.asarray(fine_labels_memmap[idx], dtype=np.int32)
            coarse_label = np.asarray(coarse_labels_memmap[idx], dtype=np.int32)
            return {"image": image, "label": label, "coarse_label": coarse_label}

        sample_spec = {
            "image": jax.ShapeDtypeStruct(shape=tuple(images_memmap.shape[1:]), dtype=jnp.uint8),
            "label": jax.ShapeDtypeStruct(shape=(), dtype=jnp.int32),
            "coarse_label": jax.ShapeDtypeStruct(shape=(), dtype=jnp.int32),
        }

        return DiskSource(
            length=int(images_memmap.shape[0]),
            sample_fn=_read_sample,
            sample_spec=sample_spec,
            ordering=ordering,
            prefetch_size=prefetch_size,
        )


"""We provide utilities for interacting with `gymnax` environments. Take care to update the policy
state in the dataloader after updating your policy parameters.

```python
import gymnax
import jax
import jax.numpy as jnp

from cyreal.transforms import BatchTransform
from cyreal.loader import DataLoader
from cyreal.rl import set_loader_policy_state, set_source_policy_state
from cyreal.sources import GymnaxSource

env = gymnax.environments.classic_control.cartpole.CartPole()
env_params = env.default_params

def act(obs, policy_state, new_episode, key):
    # policy_state can hold nn parameters and recurrent states
    # new_episode can be used to reset recurrent states
    # within the policy_state if needed.
    logits = obs @ policy_state["params"]
    action = jax.random.categorical(key, logits=logits)
    return action, policy_state

policy_state = {
    "params": jnp.zeros((4, 2)),
    "recurrent_state": jnp.zeros((3,)),
}

# GymnaxSource will call policy_step_fn to sample actions from the environment
source = GymnaxSource(
    env=env,
    env_params=env_params,
    policy_step_fn=act,
    policy_state_template=policy_state,
    # Rollouts of length 32
    steps_per_epoch=32,
)
pipeline = [
    source,
    # Rollouts are length 32, batches are length 16
    # Two batches per epoch
    BatchTransform(batch_size=16),
]
loader = DataLoader(pipeline)
state = loader.init_state(jax.random.key(0))
state = set_loader_policy_state(state, policy_state)

# Perform training
for epoch in range(2):
    for _ in range(loader.steps_per_epoch):
        batch, state, mask = jax.jit(loader.next)(state)
        # Update the rollout policy parameters after each policy update
        policy_state.update({"params": jnp.ones((4, 2))})
        state = set_loader_policy_state(state, policy_state)
```
"""



from __future__ import annotations

from pathlib import Path

from setuptools import find_packages, setup

README = (Path(__file__).parent / "README.md").read_text(encoding="utf-8")

setup(
    name="cyreal",
    version="0.1.6",
    description="Jittable data loading utilities for JAX.",
    long_description=README,
    long_description_content_type="text/markdown",
    author="",
    python_requires=">=3.10",
    packages=find_packages(exclude=("tests", "tests.*", "*.tests")),
    install_requires=[
        "jax",
        "jaxlib",
        "numpy",
    ],
    extras_require={
        "dev": ["pytest", "equinox", "optax", "tqdm", "stable-gymnax"],
    },
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



"""Jittable dataset utilities for JAX.

We use a producer and transform model. Sources stream data from datasets and Transforms transform the data stream. The DataLoader class composes sources and transforms into jittable data pipelines.


- `cyreal.datasets` contains datasets such as MNIST, CIFAR-10, etc and associated utilities.
- `cyreal.loader` contains the `DataLoader` class for building jittable data pipelines.
- `cyreal.sources` contains data sources such as `ArraySource` and `GymnaxSource`.
- `cyreal.transforms` contains data transforms such as `BatchTransform` and `DevicePutTransform`.
- `cyreal.rl` contains some RL-specific utilities for the DataLoader.
"""

# from .datasets import (
#     CIFAR10Dataset,
#     CIFAR100Dataset,
#     EMNISTDataset,
#     FashionMNISTDataset,
#     KMNISTDataset,
#     MNISTDataset,
# )
# from .loader import (
#     DataLoader,
# )
# from .transforms import (
#     BatchTransform,
#     DevicePutTransform,
#     FlattenTransform,
#     HostCallbackTransform,
#     MapTransform,
#     NormalizeImageTransform,
#     BufferTransform,
#     TimeSeriesBatchTransform,
# )



"""CelebA dataset loader with minimal dependencies."""
from __future__ import annotations

import zipfile
from dataclasses import dataclass
from pathlib import Path
from typing import Literal

import jax
import jax.numpy as jnp
import numpy as np

from .dataset_protocol import DatasetProtocol
from ..sources import DiskSource
from .utils import ensure_file as _ensure_file
from .utils import resolve_cache_dir, to_host_jax_array as _to_host_jax_array

CELEBA_URLS = {
    "images_zip": "https://s3-us-west-1.amazonaws.com/udacity-dlnfd/datasets/celeba.zip",
    "partition": "https://drive.google.com/uc?export=download&id=0B7EVK8r0v71pY0NSMzRuSXJEVkk",
    "attributes": "https://drive.google.com/uc?export=download&id=0B7EVK8r0v71pblRyaVFSWGxPY0U",
}

CELEBA_FILES = {
    "images_zip": "img_align_celeba.zip",
    "partition": "list_eval_partition.txt",
    "attributes": "list_attr_celeba.txt",
}

_SPLIT_TO_PARTITION = {"train": 0, "valid": 1, "test": 2}


def _ensure_text_file(path: Path, url: str, description: str) -> Path:
    try:
        return _ensure_file(path, url)
    except Exception as exc:
        raise FileNotFoundError(
            f"Could not download CelebA {description} from {url}."
        ) from exc


def _ensure_image_archive(path: Path, url: str) -> Path:
    if path.exists() and zipfile.is_zipfile(path):
        return path

    if path.exists():
        path.unlink()

    try:
        _ensure_file(path, url)
    except Exception as exc:
        raise FileNotFoundError(
            f"Could not download CelebA image archive from {url}."
        ) from exc

    if not zipfile.is_zipfile(path):
        path.unlink(missing_ok=True)
        raise FileNotFoundError(
            "Downloaded CelebA archive is not a valid zip file. "
            "Please provide a local `data_dir` containing `img_align_celeba/`."
        )

    return path


def _ensure_image_dir(root: Path, image_archive_url: str) -> Path:
    image_dir = root / "img_align_celeba"
    nested_dir = image_dir / "img_align_celeba"
    if image_dir.exists() and nested_dir.exists():
        return nested_dir
    if image_dir.exists():
        return image_dir

    archive_path = root / CELEBA_FILES["images_zip"]
    _ensure_image_archive(archive_path, image_archive_url)

    with zipfile.ZipFile(archive_path, "r") as archive:
        archive.extractall(path=root)

    image_dir = root / "img_align_celeba"
    nested_dir = image_dir / "img_align_celeba"
    if image_dir.exists() and nested_dir.exists():
        return nested_dir
    if image_dir.exists():
        return image_dir

    raise FileNotFoundError("Could not find `img_align_celeba/` after extraction.")


def _read_split_filenames(partition_path: Path, split: Literal["train", "valid", "test"]) -> list[str]:
    split_id = _SPLIT_TO_PARTITION[split]
    filenames: list[str] = []

    with partition_path.open("r", encoding="utf-8") as f:
        for line in f:
            parts = line.strip().split()
            if len(parts) != 2:
                continue
            name, partition = parts
            if partition.isdigit() and int(partition) == split_id:
                filenames.append(name)

    if not filenames:
        raise ValueError(f"No CelebA samples found for split '{split}'.")

    return filenames


def _read_attributes(attributes_path: Path, filenames: list[str]) -> tuple[np.ndarray, tuple[str, ...]]:
    wanted = set(filenames)
    attr_by_name: dict[str, np.ndarray] = {}

    with attributes_path.open("r", encoding="utf-8") as f:
        count_line = f.readline()
        if not count_line:
            raise ValueError("CelebA attributes file is empty.")

        names_line = f.readline().strip()
        if not names_line:
            raise ValueError("CelebA attributes header line is missing.")

        attribute_names = tuple(names_line.split())

        for line in f:
            parts = line.strip().split()
            if len(parts) < 2:
                continue
            image_name = parts[0]
            if image_name not in wanted:
                continue
            values = np.asarray(parts[1:], dtype=np.int8)
            attr_by_name[image_name] = ((values + 1) // 2).astype(np.int8)

    missing = [name for name in filenames if name not in attr_by_name]
    if missing:
        raise ValueError(f"Missing CelebA attributes for image {missing[0]}.")

    attributes = np.stack([attr_by_name[name] for name in filenames], axis=0)
    return attributes, attribute_names


def _load_image(path: Path) -> np.ndarray:
    try:
        from PIL import Image
    except ImportError as exc:  # pragma: no cover - environment-dependent
        raise ImportError("CelebA image loading requires Pillow (`pip install pillow`).") from exc

    with Image.open(path) as img:
        return np.asarray(img.convert("RGB"), dtype=np.uint8)


@dataclass
class CelebADataset(DatasetProtocol):
    """CelebA face attributes dataset."""

    split: Literal["train", "valid", "test"] = "train"
    cache_dir: str | Path | None = None
    data_dir: str | Path | None = None
    include_attributes: bool = True
    image_archive_url: str = CELEBA_URLS["images_zip"]
    partition_url: str = CELEBA_URLS["partition"]
    attributes_url: str = CELEBA_URLS["attributes"]

    def __post_init__(self) -> None:
        root = Path(self.data_dir) if self.data_dir is not None else resolve_cache_dir(
            self.cache_dir, default_name="cyreal_celeba"
        )
        root.mkdir(parents=True, exist_ok=True)

        image_dir = _ensure_image_dir(root, self.image_archive_url)
        partition_path = _ensure_text_file(root / CELEBA_FILES["partition"], self.partition_url, "partition file")

        filenames = _read_split_filenames(partition_path, self.split)
        image_paths = [image_dir / name for name in filenames]
        missing = [p.name for p in image_paths if not p.exists()]
        if missing:
            raise FileNotFoundError(f"Missing CelebA image file, e.g. {missing[0]}.")

        images_np = np.stack([_load_image(path) for path in image_paths], axis=0)
        self._images = _to_host_jax_array(images_np)

        self._attribute_names: tuple[str, ...] = tuple()
        self._attributes: jax.Array | None = None
        if self.include_attributes:
            attributes_path = _ensure_text_file(
                root / CELEBA_FILES["attributes"],
                self.attributes_url,
                "attributes file",
            )
            attributes_np, names = _read_attributes(attributes_path, filenames)
            self._attributes = _to_host_jax_array(attributes_np.astype(np.int32))
            self._attribute_names = names

    def __len__(self) -> int:
        return int(self._images.shape[0])

    def __getitem__(self, index: int):
        sample = {"image": self._images[index]}
        if self._attributes is not None:
            sample["attributes"] = self._attributes[index]
        return sample

    @property
    def attribute_names(self) -> tuple[str, ...]:
        return self._attribute_names

    def as_array_dict(self) -> dict[str, jax.Array]:
        data: dict[str, jax.Array] = {"image": self._images}
        if self._attributes is not None:
            data["attributes"] = self._attributes
        return data

    @classmethod
    def make_disk_source(
        cls,
        *,
        split: Literal["train", "valid", "test"] = "train",
        cache_dir: str | Path | None = None,
        data_dir: str | Path | None = None,
        include_attributes: bool = True,
        image_archive_url: str = CELEBA_URLS["images_zip"],
        partition_url: str = CELEBA_URLS["partition"],
        attributes_url: str = CELEBA_URLS["attributes"],
        ordering: Literal["sequential", "shuffle"] = "shuffle",
        prefetch_size: int = 64,
    ) -> DiskSource:
        root = Path(data_dir) if data_dir is not None else resolve_cache_dir(
            cache_dir, default_name="cyreal_celeba"
        )
        root.mkdir(parents=True, exist_ok=True)

        image_dir = _ensure_image_dir(root, image_archive_url)
        partition_path = _ensure_text_file(root / CELEBA_FILES["partition"], partition_url, "partition file")

        filenames = _read_split_filenames(partition_path, split)
        image_paths = [image_dir / name for name in filenames]
        missing = [p.name for p in image_paths if not p.exists()]
        if missing:
            raise FileNotFoundError(f"Missing CelebA image file, e.g. {missing[0]}.")

        first_image = _load_image(image_paths[0])
        image_shape = tuple(first_image.shape)

        attributes_np: np.ndarray | None = None
        if include_attributes:
            attributes_path = _ensure_text_file(
                root / CELEBA_FILES["attributes"],
                attributes_url,
                "attributes file",
            )
            attributes_np, _ = _read_attributes(attributes_path, filenames)

        def _read_sample(index: int | np.ndarray) -> dict[str, np.ndarray]:
            idx = int(np.asarray(index))
            sample = {"image": _load_image(image_paths[idx])}
            if attributes_np is not None:
                sample["attributes"] = np.asarray(attributes_np[idx], dtype=np.int32)
            return sample

        sample_spec: dict[str, jax.ShapeDtypeStruct] = {
            "image": jax.ShapeDtypeStruct(shape=image_shape, dtype=jnp.uint8)
        }
        if attributes_np is not None:
            sample_spec["attributes"] = jax.ShapeDtypeStruct(
                shape=(int(attributes_np.shape[1]),),
                dtype=jnp.int32,
            )

        return DiskSource(
            length=int(len(image_paths)),
            sample_fn=_read_sample,
            sample_spec=sample_spec,
            ordering=ordering,
            prefetch_size=prefetch_size,
        )
